/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <zlib.h>

#include <pspkernel.h>
#include <pspdebug.h>
#include <pspsdk.h>
#include <pspctrl.h>
#include <pspthreadman.h>

#include <pspwlan.h>
#include <pspkernel.h>
#include <psppower.h>

#include <stdlib.h>
#include <stdio.h>

#include "psp_irkeyb.h"

#include "pg.h"
#include "dialog.h"
#include "vncdbg.h"
#include "menu.h"

#define STDOUT_FILE	"stdout.txt"
#define STDERR_FILE	"stderr.txt"

/* If application's main() is redefined as SDL_main, and libSDLmain is
   linked, then this file will create the standard exit callback,
   define the PSP_* macros, add an exception handler, nullify device 
   checks and exit back to the browser when the program is finished. */

extern int SDL_main(int argc, char *argv[]);

static void cleanup_output(void);

# ifndef PSPFW30X
PSP_MODULE_INFO("PSPVNC", 0x1000, 1, 1);
# else
PSP_MODULE_INFO("PSPVNC", 0x0, 1, 1);
PSP_HEAP_SIZE_KB(12*1024);
# endif
PSP_MAIN_THREAD_ATTR(0);
PSP_MAIN_THREAD_STACK_SIZE_KB(32);

int 
sdl_psp_exit_callback(int arg1, int arg2, void *common)
{
  cleanup_output();
  sceKernelExitGame();
	return 0;
}

int sdl_psp_callback_thread(SceSize args, void *argp)
{
	int cbid;
	cbid = sceKernelCreateCallback("Exit Callback", 
				       sdl_psp_exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);

	sceKernelSleepThreadCB();
	return 0;
}

int 
sdl_psp_setup_callbacks(void)
{
	int thid = 0;

	sceCtrlSetSamplingCycle(0);
	sceCtrlSetSamplingMode(PSP_CTRL_MODE_ANALOG);

	thid = sceKernelCreateThread("update_thread", 
				     sdl_psp_callback_thread, 0x11, 0xFA0, 0, 0);
	if(thid >= 0)
		sceKernelStartThread(thid, 0, 0);
	return thid;
}

/* Remove the output files if there was no output written */
static void cleanup_output(void)
{
#ifndef NO_STDIO_REDIRECT
	FILE *file;
	int empty;
#endif

	/* Flush the output in case anything is queued */
	fclose(stdout);
	fclose(stderr);

#ifndef NO_STDIO_REDIRECT
	/* See if the files have any output in them */
	file = fopen(STDOUT_FILE, "rb");
	if ( file ) {
		empty = (fgetc(file) == EOF) ? 1 : 0;
		fclose(file);
		if ( empty ) {
			remove(STDOUT_FILE);
		}
	}
	file = fopen(STDERR_FILE, "rb");
	if ( file ) {
		empty = (fgetc(file) == EOF) ? 1 : 0;
		fclose(file);
		if ( empty ) {
			remove(STDERR_FILE);
		}
	}
#endif
# ifdef USE_PSP_IRKEYB
  psp_irkeyb_exit();
# endif
}

extern void _fini(void);

#ifdef main
#undef main
#endif

#define MAXPATH 512
#define MAXNAME 256
#define MAX_ENTRY 2048

char VNCPath[MAXPATH];
char logpath[MAXPATH];

int bSleep = 0, bQuit = 0, logcnt = 0, bDisconnected = 0;

static void InitInet();
static void ShutdownInet();
void vncDisconnect();
void vncConnect();

extern void LoadOptions();

int 
loc_main(int argc, char **argv)
{
    char *p;

    strcpy(VNCPath, argv[0]);
    p = strrchr(VNCPath, '/');
    *++p=0;

    sprintf(logpath, "%svnclog%i.txt", VNCPath, logcnt++);

    pgInit();

    PostProgressMessage("", 0, 0);

    return 0;
}

static char WiFiStateMessage[5][32] = {
  "\tWiFi Idle",
  "\tWiFi Init",
  "\tWiFi Connect",
  "\tWiFi DHCP",
  "\tWiFi Ready"
};

static void ShutdownInet() {
    int err;

    PostProgressMessage("\tStopping WiFi", 1, 4);
    err = sceNetApctlDisconnect();
    LogPn1("sceNetApctlDisconnect returns ", err, "\n");
    return;
}

extern int pickConnection(); 

static void InitInet()
{
    int net;
    int last_state;
    u32 err;
    u32 state, ostate=0, statechange=0;
    char szMyIPAddr[32];

    bDisconnected = 0;

  //LUDO:
	while (sceWlanGetSwitchState() != 1)	//switch = off
	{
    MsgPrintf(0, "\tTurn the wifi switch on !\n");
    sceKernelDelayThread(1000000);
	}

retry_net:
    net = pickConnection();
    if(net<0) {
      bQuit = 1;
      return;
    }

    // try first connection
    err = sceNetApctlConnect(net); // how to pick this ?
    LogPn1("sceNetApctlConnect returns ", err, "\n");
    if (err != 0)
        goto error_wifi;

    state = 0;
    err = sceNetApctlGetState(&state);
    LogPn1("getstate: err=", err, ", ");
    LogPn1("state=", state, "\n");
    if (err != 0)
        goto error_connection;

    last_state = -1;

    // 4 connected with IP address
    while (1)
    {
      sceKernelDelayThread(50*1000);
      err = sceNetApctlGetState(&state);

      if (err != 0)
      {
        LogPn1("sceNetApctlGetState returns ", err, "\n");
        goto close_connection;
      }

      {
        SceCtrlData c;
        sceCtrlReadBufferPositive(&c, 1);
        if (c.Buttons & PSP_CTRL_SQUARE) goto close_net;
      }


      if (state != last_state) 
      {
        if (last_state == 2 && state == 0) {
				 sceKernelDelayThread(500*1000); // 500ms
				 goto retry_net;
        }

        last_state = state;
      }

      PostProgressMessage(WiFiStateMessage[state], state, 4);

      readpad(0);
        
        // 0 - idle
        // 1,2 - starting up
        // 3 - waiting for dynamic IP -- REVIEW: wedges with DHCP
        // 4 - got IP - usable
      if (state == 4) break;  // connected with static IP
    }

    if(!bQuit) {
      // get IP address
      if (sceNetApctlGetInfo(8, szMyIPAddr) != 0)
          strcpy(szMyIPAddr, "unknown IP address");

      LogPrintf("sceNetApctlGetInfo #8: ipaddr=%s\n",szMyIPAddr);
    }

    return;

close_connection:
    PostProgressMessage("\tTerminating WiFi", 1, 4);
    err = sceNetApctlDisconnect();
    LogPn1("sceNetApctlDisconnect returns ", err, "\n");
    goto retry_net;

error_timeout:
    MsgPrintf(0, "\tWiFi Connection Timed Out\n\n\nPlease check the WiFi switch\n\nand try again.");
    goto close_connection;

error_connection:
    MsgPrintf(0, "\tAn error occured while connecting\n\nto the access point.\n\n\nPlease check your settings\n\nand try again.");
    goto close_connection;

error_wifi:
    MsgPrintf(0, "\tError Connecting to Network\n\n\nPlease check your settings\n\nand try again.");
    goto retry_net;

close_net:
  err = sceNetApctlDisconnect();

  nlhTerm();
  sceKernelDelayThread(1000000); 
}



void
user_thread(SceSize args, void *argp)
{
  if (! danzeff_load()) {
    fprintf(stdout, "failed to load virtual keyboard\n");
    return;
  }

  int err = nlhInit();
  LogPn1("nlhInit returns ", err, "\n");
  if (err != 0) bQuit=1;

  while(!bQuit) {
    InitInet();
    vncMenu();
    ShutdownInet();
  }
 
  PostProgressMessage("\tStopping NetLib", 2, 4);
  nlhTerm();
  sceKernelDelayThread(1000000); 
}

int 
main(int argc, char *argv[])
{
  int user_thid;

	pspDebugScreenInit();

# ifdef PSPFW30X
  pspDebugScreenPrintf("PSPVNC v1.2.7-3X0E: Portable VNC Client\n");
# else
  pspDebugScreenPrintf("PSPVNC v1.2.7-15SE: Portable VNC Client\n");
# endif
  pspDebugScreenPrintf("Please wait now loading !\n");
  sceKernelDelayThread(1500000); 

#ifndef NO_STDIO_REDIRECT
	/* Redirect standard output and standard error. */
	/* TODO: Error checking. */
	freopen(STDOUT_FILE, "w", stdout);
	freopen(STDERR_FILE, "w", stderr);
	setvbuf(stdout, NULL, _IOLBF, BUFSIZ);	/* Line buffered */
	setbuf(stderr, NULL);					/* No buffering */
#endif /* NO_STDIO_REDIRECT */

	/* Functions registered with atexit() are called in reverse order, so make
	   sure that we register sceKernelExitGame() first, so that it's called last. */
	atexit(sceKernelExitGame);
	/* Make sure that _fini() is called before returning to the OS. */
	atexit(_fini);
	atexit(cleanup_output);

  if (nlhLoadDrivers(&module_info) != 0) {
    pspDebugScreenPrintf("unable to load wifi drivers !");
    return 0;
  }

# ifdef USE_PSP_IRKEYB
  if (psp_irkeyb_init()) {
    fprintf(stdout, "failed to initialize IR keyboard !\n");
  }
# endif

  loc_main(argc, argv);

  //pspDebugScreenPrintf("\nLaunch main thread !\n");
  //sceKernelDelayThread(1000000); 

  user_thid = sceKernelCreateThread( "user_thread", 
     (SceKernelThreadEntry)user_thread, 0x18, 256*1024, PSP_THREAD_ATTR_USER, 0 );
  if(user_thid >= 0) {
    sceKernelStartThread(user_thid, 0, 0);
    sceKernelWaitThreadEnd(user_thid, NULL);
  }

	/* Delay 2.5 seconds before returning to the OS. */
	sceKernelDelayThread(2500000);

	return 0;
}
